""" handle parameters like transforming, merging, converting"""
import configparser
from collections import OrderedDict

import mdpeditor.mdpblocks.provide as provide
from mdpeditor.mdpblocks.process import apply_instructions


def compile_input_token_to_parameters(word):

    try:
        return provide.mdp_string_to_ordered_dict(word)
    except configparser.ParsingError:
        pass

    try:
        return provide.read_parameter_block(word)
    # depending on the word format we'll fail at different positions
    # if this is not a well formed parameter block name a.b
    except ModuleNotFoundError:
        # has form a.b , where a does not exist as a module
        pass
    except FileNotFoundError:
        # has form a.b where a exists, but not b
        pass
    except IndexError:
        # does not have form a.b
        pass

    try:
        with open(word) as f:
            mdp_input_as_string = f.read()
        return provide.mdp_string_to_ordered_dict(mdp_input_as_string)
    except FileNotFoundError:
        pass
    except configparser.ParsingError:
        pass

    raise ValueError(f"{word} cannot be read as an .mdp file, a block name or "
                     f"an attempt to set an .mdp option.")


def compile_parameters(compile_input):

    parameter_blocks = [
        compile_input_token_to_parameters(token) for token in compile_input
    ]

    parameters = OrderedDict()
    duplicate_keys = set()

    for block in parameter_blocks:

        # collect overlapping parameters in the blocks
        duplicate_keys |= set(parameters.keys()) & set(block.keys())
        parameters.update(block)

    apply_instructions(parameters, compile_input)

    return OutputParameters(parameters, duplicate_keys)


class OutputParameters:
    """
    Handle parameter output like printing and cleaning
    """
    def __init__(self, parameters, duplicate_keys):

        # open default .mdp and fill in the "other parameters"
        # discard all non-default parameters
        self.parameters = provide.default_parameter_block()
        for key in self.parameters.keys():
            if key in parameters.keys():
                self.parameters[key] = parameters[key]

        self.duplicate_keys = duplicate_keys
        self.modified_keys = parameters.keys()

    def keep_only_modified(self):
        """ Remove all parameters that have not been modified """

        keys_to_pop = [
            key for key in self.parameters.keys()
            if key not in self.modified_keys
        ]
        for key in keys_to_pop:
            self.parameters.pop(key)

    def as_string(self):
        """ write the parameters as a string in .mdp format"""

        if not self.parameters:
            return ""

        modifed_key_style = "bold"

        max_key_length = max(map(len, self.parameters.keys()))

        formatted_mdp_entries = [
            f'[{modifed_key_style}]{key:{max_key_length}s} = '
            f'{value}[/{modifed_key_style}]'
            if key in self.modified_keys else
            f'{key:{max_key_length}s} = {value}'
            for key, value in self.parameters.items()
        ]

        return '\n'.join(formatted_mdp_entries) + "\n"



import argparse
import sys

from mdpeditor.mdpblocks import provide


def add_explain_subparser(subparsers):
    explain_parser = subparsers.add_parser("explain")

    explain_parser.add_argument(
        dest="explain",
        nargs='?',
        metavar="explain",
        help="explain an .mdp parameter or parameter block",
    )


def add_compile_subparser(subparsers):

    compile_parser = subparsers.add_parser("compile")

    compile_parser.add_argument(dest="compile",
                                nargs="*",
                                metavar="compile tokens")

    compile_parser.add_argument(
        "--merge-duplicates",
        dest="merge_right",
        action='store_true',
        default=False,
        help="allow duplicate parameters by overwriting previously set"
        " parameters")

    compile_parser.add_argument(
        "--output",
        nargs='?',
        const='compiled.mdp',
        type=argparse.FileType('w'),
        help="write the compiled parameters to an .mdp file"
        " (instead of command line)",
        metavar="compiled.mdp")

    compile_parser.add_argument("--full-mdp",
                                dest="full_mdp",
                                action='store_true',
                                default=False,
                                help="write all mdp options"
                                "including defaults")


def get_command_line_arguments(version: str):
    """build, parse and return command line arguments

    Args:
        version (str): the current program version

    Returns:
        the parsed command line arguments
    """

    program_name = "mdpeditor"
    description = (
        "Compiles an .mdp file from preset .mdp parameter blocks"
        " and user settings. To learn more about available parameters"
        " use explain. ")

    epilog = """examples:
    mdpeditor explain
    \tShows available pre-defined parameter blocks

    mdpeditor compile force_field.charmm nsteps=100
    \tCompiles the pre-defined block force_field.charmm and
    \tsets "nsteps = 100" in the output
    """

    parser = argparse.ArgumentParser(
        description=description,
        prog=program_name,
        add_help=False,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    subparsers = parser.add_subparsers(help='', dest='command')

    parser.add_argument("-h",
                        "--help",
                        action="help",
                        help="show this help message and exit")

    parser.add_argument("--version",
                        action="version",
                        version=(f"{program_name} {version}"))

    add_compile_subparser(subparsers)
    add_explain_subparser(subparsers)

    return parser.parse_args()



""" Run the command line interface """
import argparse
import importlib.metadata
import itertools
import sys

from mdpeditor.explain import run_explain
import mdpeditor.compile
import mdpeditor.commandlineinterface.prompt
import mdpeditor.commandlineinterface.arguments as arguments

from rich.console import Console
from rich import print


def run():
    """ run the command line interface """

    # set up the console for printing
    console = Console()

    # derive the program version via git
    try:
        version = importlib.metadata.version("mdpeditor")
    except importlib.metadata.PackageNotFoundError:
        version = "Unknown"

    command_line_arguments = arguments.get_command_line_arguments(version)

    if command_line_arguments.command == "compile":
        # print some hints if the input string is empty
        if not command_line_arguments.compile:
            console.print(mdpeditor.compile.help_string())
            sys.exit()

        output_string = mdpeditor.compile.run_compile(
            command_line_arguments.compile, command_line_arguments.merge_right,
            command_line_arguments.full_mdp)

        mdpeditor.compile.print_annotated_output(console, output_string,
                                                 version, sys.argv,
                                                 command_line_arguments.output)

        sys.exit()

    if command_line_arguments.command == "explain":
        try:
            run_explain(console, command_line_arguments.explain)
        except SystemExit as e:
            console.print(e.__str__())
            raise SystemExit("")
        sys.exit()

    if not command_line_arguments.command:
        mdpeditor.commandlineinterface.prompt.run_interactive_prompt(
            console, version)



import rich.markdown
import mdpeditor.mdpblocks.provide as provide
from mdpeditor.parameterhelp import extract


def run_explain(console, explain_string):

    if not explain_string:
        console.print("\nUse as \n\texplain [italic]PARAMETER \n[/]"
                      "where [italic]PARAMETER[/] is "
                      "one of these predefined parameter blocks\n")
        block_names = '\n\t'.join(provide.available_parameter_blocks())
        console.print("[bold]\t" + block_names + "[/]")
        console.print("\nor an .mdp option like [bold]integrator[/].\n")
        return

    explain_string = explain_string.strip()
    try:
        description_as_md = rich.markdown.Markdown(
            extract.mdp_section(explain_string))
        console.print(description_as_md)
        return
    except ValueError as e:
        pass

    try:
        plain_string = provide.description_string(explain_string)
        as_markdown = rich.markdown.Markdown(plain_string)
        console.print(as_markdown)
        return
    except FileNotFoundError as e:
        pass
    except IndexError:
        pass
    except ModuleNotFoundError:
        pass

    raise SystemExit(f"[red bold]{explain_string}[/] is neither an "
                     f".mdp option nor a predefined block")



import readline
import mdpeditor.mdpblocks.provide
import mdpeditor.compile
import mdpeditor.explain
import mdpeditor.parameterhelp.extract


class VolcabCompleter:
    def __init__(self, volcab):
        self.volcab = volcab

    def complete(self, text, state):
        results = [x + " " for x in self.volcab if x.startswith(text)] + [None]
        return results[state]


def console_prompt_string(compile_mode):
    modes = ["compile", "explain"]

    if compile_mode:
        modes.reverse()

    console_prompt_string = ("\nPress [bold][ENTER][/bold] to switch to " +
                             f"{modes[0]}" + " mode.\n")
    console_prompt_string += (
        "Press [bold][TAB][/bold] to list input options.\n")
    console_prompt_string += ("Press [bold][CTRL]+C[/bold] to exit.\n")

    console_prompt_string += f"\n> [bold]{modes[1]}[/] "

    if compile_mode:
        console_prompt_string += "[italic]--merge-duplicates[/] "

    return console_prompt_string


def run_interactive_prompt(console, version):

    error_style = "bold color(1)"
    simple_text_style = "color(6)"

    # introductory message
    console.rule(f"mdpeditor {version}", style="")
    console.print(
        "Welcome to the interactive mode of mdpeditor!"
        "\n\nHere you"
        " can learn about .mdp parameters and parameter blocks "
        "and test different parameter combinations.\n\nFor production"
        " code in workflows, use the command line options as shown"
        " when running [italic]mdpeditor --help.\n",
        style=simple_text_style)
    available_blocks = mdpeditor.mdpblocks.provide.available_parameter_blocks()
    console.print("To help you compile your .mdp file, you can currently "
                  "choose from these blocks:\n\n[bold]" +
                  "\n".join(available_blocks),
                  style=simple_text_style)
    console.print(
        "\nLearn more about these blocks with the [bold]explain[/] mode.\n",
        style=simple_text_style)

    console.print(
        "\nA typical input is \n"
        "[italic]compile --merge-duplicates force_field.amber "
        "pressure.atomspheric[/]\n",
        style=simple_text_style)

    # set up tab completion
    readline.parse_and_bind("tab: complete")
    readline.set_completer_delims(" ")
    blocks = mdpeditor.mdpblocks.provide.available_parameter_blocks()
    mdp_options = mdpeditor.parameterhelp.extract.mdp_options_list()
    completer = VolcabCompleter(blocks + mdp_options)
    readline.set_completer(completer.complete)

    merge_right = True
    full_mdp = False

    line = ""
    compile_mode = False
    while line is not None or line == "end":

        # swap mode if only [Enter] was pressed
        if line == "":
            compile_mode = not compile_mode

        try:
            console.rule(style="")
            line = console.input(console_prompt_string(compile_mode))
            if compile_mode:
                if line == "":
                    continue
                output = mdpeditor.compile.run_compile(line.split(),
                                                       merge_right, full_mdp)
                console.print(output)
            else:
                if line == "":
                    continue
                mdpeditor.explain.run_explain(console, line)

        # end gracefully when the user interrupts
        except KeyboardInterrupt:
            line = None
        except EOFError:
            line = None
        # do not exit like we do when running pure command line mode
        except SystemExit as e:
            console.print("Error: " + e.__str__() + "\n", style=error_style)

    console.print("\n\nThanks for using mdpeditor!")

    console.print(
        "\nDiscuss .mdp parameters at "
        "https://gromacs.bioexcel.eu/tag/mdp-parameters", justify="right"
    )

    console.print(
        "Report issues and suggestions for mdpeditor at "
        "https://gitlab.com/cblau/mdpeditor/-/issues",
        justify="right")

    console.print("\n:Copyright: 2021,2022  Christian Blau", justify="right")



import mdpeditor.parameterhandler as parameterhandler


def help_string():
    return """
Use like\n
    [bold]compile density_guided.vanilla pixel-size-in-nm=0.98 some.mdp[/]

Run [bold]explain[/] to learn more about available options"""


def run_compile(compile_command, merge_right, full_mdp):

    try:
        output_parameters = parameterhandler.compile_parameters(
            compile_command)
    except ValueError as e:
        raise SystemExit(e.__str__())
    except AttributeError as e:
        raise SystemExit()

    if (not merge_right and output_parameters.duplicate_keys):
        raise SystemExit(
            "\nAborting compilation due to duplicate parameter(s)\n\n\t" +
            "\n\t".join(list(output_parameters.duplicate_keys)) +
            "\n\nUse --merge-duplicates to override parameters\n")

    if not full_mdp:
        # discard all parameters that were not explicitely chosen
        output_parameters.keep_only_modified()

    return output_parameters.as_string()


def print_annotated_output(console, output_string, version, arguments, output):
    # keep track of the command used to generate
    # the output by prepending a commented line
    output_string = f"; Created by mdpeditor version {version}\n" + \
                    "; " + ' '.join(arguments) + \
                    "\n; To create self-documenting workflows, "\
                    "do not edit this file but rerun this tool with "\
                    "different settings\n" + output_string

    if output:
        console = Console(file=output)

    console.print(output_string, style="")
